use std::collections::HashMap;

use color_eyre::eyre;

use crate::lib::{IterExt, Counter};


type Pairs = Counter<(char, char)>;
type Rules = HashMap<(char, char), char>;


fn load(data: &str) -> eyre::Result<(Pairs, Rules)> {
    let mut lines = data.lines();
    let template = lines.next().unwrap().to_string();
    let mut pairs = Counter::new();
    for i in 1..template.len() {
        let a = template.as_bytes()[i - 1] as char;
        let b = template.as_bytes()[i] as char;
        pairs.inc((a, b));
    }
    lines.next();

    let mut rules = HashMap::new();
    for line in &mut lines {
        let (pair_s, tgt_s) = line.split(" -> ").take_pair()?;
        let pair = pair_s.chars().take_pair()?;
        let tgt = tgt_s.chars().next().unwrap();
        rules.insert(pair, tgt);
    }

    Ok((pairs, rules))
}


fn expand(src: &Pairs, rules: &Rules) -> Pairs {
    let mut dst = Counter::new();
    for (&(a, b), count) in src.iter() {
        if let Some(&tween) = rules.get(&(a, b)) {
            dst.inc_by((a, tween), count);
            dst.inc_by((tween, b), count);
        }
    }
    dst
}


fn score(pairs: &Pairs) -> usize {
    // "score" is defined as the number of occurences of the most common
    // character minus the number of occurrences of the least common
    let mut firsts = Counter::new();
    let mut lasts = Counter::new();
    for (pair, count) in pairs.iter() {
        let (a, b) = *pair;
        firsts.inc_by(a, count);
        lasts.inc_by(b, count);
    }

    // the count of a given character is the number of times it appears
    // in either first or last place of a pair, whichever is greater
    let mut counts = Counter::new();
    for (c, count) in firsts.iter().chain(lasts.iter()) {
        if count > counts.get(c) {
            counts.set(*c, count);
        }
    }

    let max = counts.iter().map(|(_c, n)| n).max().unwrap();
    let min = counts.iter().map(|(_c, n)| n).min().unwrap();
    max - min
}


pub fn run(data: &str) -> eyre::Result<(usize, usize)> {
    let (mut pairs, rules) = load(data)?;
    
    for _i in 0..10 {pairs = expand(&pairs, &rules);}
    let one = score(&pairs);

    for _i in 0..30 {pairs = expand(&pairs, &rules);}
    let two = score(&pairs);

    Ok((one, two))
}