use color_eyre::eyre;

use crate::lib::{BitVec, BitSlice};


#[derive(Debug)]
struct Packet {
    version: u8,
    type_id: u8,
    content: PacketContent,
}

#[derive(Debug)]
enum PacketContent {
    Literal(u64),
    Operator(Vec<Packet>),
}

struct Parser {
    data: BitVec,
    idx: usize,
}

impl Parser {
    fn from_hex(hex: &str) -> eyre::Result<Self> {
        let mut data = BitVec::with_capacity(hex.len() * 4);
        for i in 0 .. (hex.len() / 2) { // we are assuming valid hex
            let chunk = &hex[i * 2 .. (i + 1) * 2];
            let byte = u8::from_str_radix(chunk, 16)?;
            data.push_byte(byte);
        }

        Ok(Parser {data, idx: 0})
    }

    fn next(&mut self, n: usize) -> BitSlice {
        let range = self.idx .. self.idx + n;
        self.idx += n;
        self.data.slice(range)
    }

    fn parse_value(&mut self) -> u64 {
        let mut value = 0;
        loop {
            let group = self.next(5);
            value = (value << 4) | group.slice(1..5).as_u64();
            if !group[0] {break;}
        }
        value
    }

    fn parse_children(&mut self) -> Vec<Packet> {
        let mut children = Vec::new();
        if self.next(1).as_bool() {
            // count refers to number of sub-packets
            let count = self.next(11).as_u16();
            children = Vec::with_capacity(count as usize);
            for _i in 0..count {
                children.push(self.parse());
            }
        }
        else {
            // len refers to content length in bits
            let len = self.next(15).as_u16();
            let end = self.idx + (len as usize);
            while self.idx < end {
                children.push(self.parse());
            }
        }
        children
    }

    fn parse(&mut self) -> Packet {
        let version = self.next(3).as_u8();
        let type_id = self.next(3).as_u8();
        let content = match type_id {
            4 => PacketContent::Literal(self.parse_value()),
            _ => PacketContent::Operator(self.parse_children()),
        };

        Packet {version, type_id, content}
    }
}


fn sum_versions(root: &Packet) -> usize {
    let mut total = root.version as usize;
    if let PacketContent::Operator(children) = &root.content {
        for child in children {
            total += sum_versions(child);
        }
    }
    total
}


fn evaluate(packet: &Packet) -> usize {
    let children = match &packet.content {
        PacketContent::Literal(value) => {return *value as usize;}, // early exit here
        PacketContent::Operator(children) => children,
    };

    let mut values = children.iter().map(|p| evaluate(p));
    match packet.type_id {
        0 => values.sum(),
        1 => values.product(),
        2 => values.min().unwrap(),
        3 => values.max().unwrap(),
        // it's guaranteed that comparisons will have exactly two children
        5 => if values.next().unwrap() > values.next().unwrap() {1} else {0},
        6 => if values.next().unwrap() < values.next().unwrap() {1} else {0},
        7 => if values.next().unwrap() == values.next().unwrap() {1} else {0},
        // the version number was a 3-bit value, so:
        _ => unreachable!(),
    }
}


pub fn run(data: &str) -> eyre::Result<(usize, usize)> {
    let mut parser = Parser::from_hex(data)?;
    let root = parser.parse();

    let one = sum_versions(&root);
    let two = evaluate(&root);

    Ok((one, two))
}
