use std::str::FromStr;

use color_eyre::eyre;
use num::integer::gcd;

use crate::lib::{ParseLines, IterExt};


#[derive(Copy, Clone, Debug, Eq, PartialEq)]
struct Point {
    x: isize,
    y: isize,
}


#[derive(Copy, Clone, Debug)]
struct Slope {
    rise: isize,
    run: isize,
}

impl Slope {
    fn apply_to(&self, pt: Point) -> Point {
        Point {
            x: pt.x + self.run,
            y: pt.y + self.rise,
        }
    }
}


#[derive(Copy, Clone, Debug)]
struct Line {
    start: Point,
    end: Point,
}

impl Line {
    fn slope(&self) -> Slope {
        let dx = self.end.y - self.start.y;
        let dy = self.end.x - self.start.x;

        let divisor;
        if dx == 0 || dy == 0 {
            divisor = (dx + dy).abs();
        }
        else {
            divisor = gcd(dx, dy);
        }

        Slope {
            rise: dx / divisor,
            run: dy / divisor,
        }
    }

    fn points(&self) -> PointsIter {
        PointsIter {
            line: self,
            pos: self.start,
            slope: self.slope(),
        }
    }
}


#[derive(Debug)]
struct PointsIter<'a> {
    line: &'a Line,
    pos: Point,
    slope: Slope,
}

impl Iterator for PointsIter<'_> {
    type Item = Point;
    fn next(&mut self) -> Option<Self::Item> {
        // note, this approach will break if total dX and dY are not multiples of rise/run
        if self.pos == self.slope.apply_to(self.line.end) {
            return None;
        }

        let orig = self.pos;
        self.pos = self.slope.apply_to(self.pos);
        Some(orig)
    }
}


impl FromStr for Line {
    type Err = eyre::Report;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let (p1, p2) = s.split(" -> ").take_pair()?;
        let (x1, y1) = p1.split(',').take_pair()?;
        let (x2, y2) = p2.split(',').take_pair()?;
        
        let start = Point {
            x: x1.parse::<isize>()?,
            y: y1.parse::<isize>()?,
        };
        let end = Point {
            x: x2.parse::<isize>()?,
            y: y2.parse::<isize>()?,
        };

        Ok(Line {start, end})
    }
}


fn part1(lines: &[Line]) -> usize {
    let mut grid = vec![[0; 1000]; 1000];

    let points = lines.iter()
        .filter(|l| l.start.x == l.end.x || l.start.y == l.end.y)
        .flat_map(|l| l.points());

    for pt in points {
        grid[pt.y as usize][pt.x as usize] += 1;
    }

    grid.into_iter()
        .flat_map(|row| row.into_iter())
        .filter(|n| *n > 1)
        .count()
}


fn part2(lines: &[Line]) -> usize {
    let mut grid = vec![[0; 1000]; 1000];

    for pt in lines.iter().flat_map(|l| l.points()) {
        grid[pt.y as usize][pt.x as usize] += 1;
    }

    grid.into_iter()
        .flat_map(|row| row.into_iter())
        .filter(|n| *n > 1)
        .count()
}


pub fn run(data: &str) -> eyre::Result<(usize, usize)> {
    let mut lines = Vec::new();
    for line in data.parse_lines::<Line>() {
        lines.push(line?);
    }

    Ok((part1(&lines), part2(&lines)))
}
