use std::str::FromStr;

use color_eyre::eyre;
use eyre::eyre;

use crate::lib::{ParseLines, IterExt};


#[derive(Copy, Clone, Debug, Eq, PartialEq)]
enum Segment {
    A = 1,
    B = 2,
    C = 4,
    D = 8,
    E = 16,
    F = 32,
    G = 64, // wheee!
}
use Segment::*;

impl Segment {
    fn from_char(c: char) -> eyre::Result<Self> {
        match c {
            'a' => Ok(A),
            'b' => Ok(B),
            'c' => Ok(C),
            'd' => Ok(D),
            'e' => Ok(E),
            'f' => Ok(F),
            'g' => Ok(G),
            _ => Err(eyre!("Invalid character for segment: {}", c)),
        }
    }

    fn to_ord(&self) -> usize {
        // kinda gross but whatever
        match self {
            A => 0, B => 1, C => 2, D => 3,
            E => 4, F => 5, G => 6
        }
    }

    fn from_ord(n: u8) -> eyre::Result<Self> {
        match n {
            0 => Ok(A),
            1 => Ok(B),
            2 => Ok(C),
            3 => Ok(D),
            4 => Ok(E),
            5 => Ok(F),
            6 => Ok(G),
            _ => Err(eyre!("Invalid ordinal for segment: {}", n)),
        }
    }

    fn from_value(n: u8) -> eyre::Result<Self> {
        match n {
            1 => Ok(A),
            2 => Ok(B),
            4 => Ok(C),
            8 => Ok(D),
            16 => Ok(E),
            32 => Ok(F),
            64 => Ok(G),
            _ => Err(eyre!("Invalid value for segment: {}", n)),
        }
    }
}


#[derive(Copy, Clone, Default, Debug)]
struct SegmentSet {
    data: u8,
}

impl SegmentSet {
    fn contains(&self, segment: Segment) -> bool {
        self.data & (segment as u8) > 0
    }

    fn insert(&mut self, segment: Segment) {
        self.data |= segment as u8;
    }

    fn len(&self) -> u8 {
        (0..7).fold(0, |total, i| total + ((self.data >> i) & 1))
    }

    fn exclusion(&self, other: &Self) -> Self {
        SegmentSet {data: self.data ^ other.data}
    }

    fn iter(&self) -> SegmentSetIter {
        SegmentSetIter {ss: self, pos: 0}
    }
}

impl FromStr for SegmentSet {
    type Err = eyre::Report;
    fn from_str(s: &str) -> eyre::Result<Self> {
        let mut set = Self::default();
        for c in s.chars() {
            set.insert(Segment::from_char(c)?);
        }
        Ok(set)
    }
}


struct SegmentSetIter<'a> {
    ss: &'a SegmentSet,
    pos: u8,
}

impl Iterator for SegmentSetIter<'_> {
    type Item = Segment;
    fn next(&mut self) -> Option<Segment> {
        while self.pos < 7 {
            let seg_value = self.ss.data & (1 << self.pos);
            self.pos += 1;
            if seg_value > 0 {
                return Some(Segment::from_value(seg_value).unwrap());
            }
        }
        return None
    }
}


#[derive(Default, Debug)]
struct Screen {
    patterns: [SegmentSet; 10],
    output: [SegmentSet; 4],
}

impl FromStr for Screen {
    type Err = eyre::Report;
    fn from_str(s: &str) -> eyre::Result<Self> {
        let mut screen = Screen::default();
        let (patterns, output) = s.split(" | ").take_pair()?;
        
        for (i, patt) in patterns.split(' ').enumerate() {
            screen.patterns[i] = patt.parse::<SegmentSet>()?;
        }
        for (i, out) in output.split(' ').enumerate() {
            screen.output[i] = out.parse::<SegmentSet>()?;
        }
        Ok(screen)
    }
}


fn segment_counts(screen: &Screen) -> [usize; 7] {
    let mut counts = [0; 7];
    for i in 0..7 {
        let segment = Segment::from_ord(i).unwrap();
        counts[i as usize] = screen.patterns.iter()
            .filter(|p| p.contains(segment))
            .count()
    }
    counts
}


fn find_a(screen: &Screen) -> Segment {
    // figure out which segment is currently attached to the "A" wire
    let mut seven = None;
    let mut one = None;
    for p in screen.patterns {
        if p.len() == 2 {
            one = Some(p); // the only digit that uses 2 segments is 1
        }
        else if p.len() == 3 {
            seven = Some(p); // similarly for 7
        }
    }

    let result = seven.unwrap().exclusion(&one.unwrap());
    if result.len() != 1 {
        panic!("This should never happen");
    }
    result.iter().next().unwrap()
}


fn solve(screen: &Screen) -> [Segment; 7] {
    let mut result = [A; 7]; // A is just a placeholder value
    let four_pattern = screen.patterns.iter().find(|p| p.len() == 4).unwrap();
    for (i, count) in segment_counts(screen).iter().enumerate() {
        let possible = match count {
            4 => Some(E),
            6 => Some(B),
            7|8 => None,
            9 => Some(F),
            _ => unreachable!(),
        };

        if let Some(segment) = possible {
            result[i] = segment;
        }
        else if *count == 7 { // 7 appearances means either D or G
            let segment = Segment::from_ord(i as u8).unwrap();
            if four_pattern.contains(segment) {
                result[i] = D;
            }
            else {
                result[i] = G;
            }
        }
        else if *count == 8 { // similarly 8 should be either A or C
            let segment = Segment::from_ord(i as u8).unwrap();
            if segment == find_a(screen) {
                result[i] = A;
            }
            else {
                result[i] = C;
            }
        }
    }
    result
}


fn decode(output: [SegmentSet; 4], map: [Segment; 7]) -> usize {
    let mut result = 0;
    for (i, segments_enc) in output.iter().enumerate() {
        let mut segments_dec = SegmentSet::default();
        for segment_enc in segments_enc.iter() {
            let segment_dec = map[segment_enc.to_ord()];
            segments_dec.insert(segment_dec)
        }
        
        let digit = match segments_dec.data {
            0b01110111 => 0,
            0b00100100 => 1,
            0b01011101 => 2,
            0b01101101 => 3,
            0b00101110 => 4,
            0b01101011 => 5,
            0b01111011 => 6,
            0b00100101 => 7,
            0b01111111 => 8,
            0b01101111 => 9,
            _ => {
                dbg!(segments_dec);
                panic!("Invalid segment sequence");
            },
        };

        result += digit * 10usize.pow(3 - (i as u32));
    }
    result
}


fn part2(screens: &[Screen]) -> usize {
    let mut result = 0;
    for screen in screens {
        let map = solve(screen);
        result += decode(screen.output, map);
    }
    result
}


fn part1(screens: &[Screen]) -> usize {
    screens.iter()
        .flat_map(|d| d.output.iter())
        .filter(|s| matches!(s.len(), 2|3|4|7))
        .count()
}


pub fn run(data: &str) -> eyre::Result<(usize, usize)> {
    let screens = data
        .parse_lines::<Screen>()
        .collect::<Result<Vec<_>, _>>()?;

    Ok((part1(&screens), part2(&screens)))
}
